import numpy as np
import datatable as dt
from scipy.sparse import issparse, csr_matrix, find, coo_matrix
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.manifold.t_sne import _joint_probabilities, _joint_probabilities_nn
from sklearn.decomposition import PCA
from scipy.spatial.distance import squareform
from sklearn.neighbors import NearestNeighbors
from scipy.sparse.linalg import eigs
from numpy.linalg import norm


def magic(data=None, aff_mat_input=0, n_pca_components=20, n_diffusion_components = 10, random_pca=True, 
          k=30, ka=10, epsilon=1, csv_l=None, csv_d=None):

    if data != None:
        print('doing PCA')
        affin_mat_input = run_pca(data, n_components=n_pca_components, random=random_pca)
    else:
        print('using user-supplied dimensional reduction')
        affin_mat_input = aff_mat_input

    #run diffusion maps to get markov matrix
    L = compute_markov(affin_mat_input, k=k, epsilon=epsilon, 
                       distance_metric='euclidean', ka=ka)
    L = coo_matrix(L)
    L = dt.Frame(row=L.row, col=L.col, data=L.data)
    L.to_csv(csv_l)

    if n_diffusion_components != 0:                
        # Eigen value decomposition
        D, V = eigs(L, k=n_diffusion_components, tol=1e-4, maxiter=1000)
        D = np.real(D)
        V = np.real(V)
        inds = np.argsort(D)[::-1]
        D = D[inds]
        V = V[:, inds]
        V = L.dot(V)
        # Normalize
        for i in range(V.shape[1]):
            V[:, i] = V[:, i] / norm(V[:, i])
        V = np.round(V, 10)
        V.to_csv(csv_d)
    #remove tsne kernel for now
    # else:
    #     distances = pairwise_distances(pca_projected_data, squared=True)
    #     if k_knn > 0:
    #         neighbors_nn = np.argsort(distances, axis=0)[:, :k_knn]
    #         P = _joint_probabilities_nn(distances, neighbors_nn, perplexity, 1)
    #     else:
    #         P = _joint_probabilities(distances, perplexity, 1)
    #     P = squareform(P)

    #     #markov normalize P
    #     L = np.divide(P, np.sum(P, axis=1))

    #get imputed data matrix -- by default use data_norm but give user option to pick
    #new_data, L_t = impute_fast(data, L, t, rescale_percent=rescale)



def run_pca(data, n_components=100, random=True):

    solver = 'randomized'
    if random != True:
        solver = 'full'

    pca = PCA(n_components=n_components, svd_solver=solver)
    return pca.fit_transform(data)


def impute_fast(data, L, t, rescale_percent=0, L_t=None, tprev=None):

    #convert L to full matrix
    if issparse(L):
        L = L.todense()

    #L^t
    print('MAGIC: L_t = L^t')
    if L_t == None:
        L_t = np.linalg.matrix_power(L, t)
    else:
        L_t = np.dot(L_t, np.linalg.matrix_power(L, t-tprev))

    print('MAGIC: data_new = L_t * data')
    data_new = np.array(np.dot(L_t, data))

    #rescale data
    if rescale_percent != 0:
        if len(np.where(data_new < 0)[0]) > 0:
            print('Rescaling should not be performed on log-transformed '
                  '(or other negative) values. Imputed data returned unscaled.')
            return data_new, L_t
            
        M99 = np.percentile(data, rescale_percent, axis=0)
        M100 = data.max(axis=0)
        indices = np.where(M99 == 0)[0]
        M99[indices] = M100[indices]
        M99_new = np.percentile(data_new, rescale_percent, axis=0)
        M100_new = data_new.max(axis=0)
        indices = np.where(M99_new == 0)[0]
        M99_new[indices] = M100_new[indices]
        max_ratio = np.divide(M99, M99_new)
        data_new = np.multiply(data_new, np.tile(max_ratio, (len(data), 1)))
    
    return data_new, L_t


def compute_markov(data, k=10, epsilon=1, distance_metric='euclidean', ka=0):

    N = data.shape[0]

    # Nearest neighbors
    print('Computing distances')
    nbrs = NearestNeighbors(n_neighbors=k, metric=distance_metric).fit(data)
    distances, indices = nbrs.kneighbors(data)

    if ka > 0:
        print('Autotuning distances')
        for j in reversed(range(N)):
            temp = sorted(distances[j])
            lMaxTempIdxs = min(ka, len(temp))
            if lMaxTempIdxs == 0 or temp[lMaxTempIdxs] == 0:
                distances[j] = 0
            else:
                distances[j] = np.divide(distances[j], temp[lMaxTempIdxs])

    # Adjacency matrix
    print('Computing kernel')
    rows = np.zeros(N * k, dtype=np.int32)
    cols = np.zeros(N * k, dtype=np.int32)
    dists = np.zeros(N * k)
    location = 0
    for i in range(N):
        inds = range(location, location + k)
        rows[inds] = indices[i, :]
        cols[inds] = i
        dists[inds] = distances[i, :]
        location += k
    if epsilon > 0:
        Wd = csr_matrix( (dists, (rows, cols)), shape=[N, N] )
    else:
        Wd = csr_matrix( (np.ones(dists.shape), (rows, cols)), shape=[N, N] )

    # Symmetrize W
    W = Wd + Wd.T

    if epsilon > 0:
        # Convert to affinity (with selfloops)
        rows, cols, dists = find(W)
        rows = np.append(rows, range(N))
        cols = np.append(cols, range(N))
        dists = np.append(dists/(epsilon ** 2), np.zeros(N))
        W = csr_matrix( (np.exp(-dists), (rows, cols)), shape=[N, N] )

    # Create D
    D = np.ravel(W.sum(axis = 1))
    D[D!=0] = 1/D[D!=0]

    #markov normalization
    T = csr_matrix((D, (range(N), range(N))), shape=[N, N]).dot(W)

    return T


def optimal_t(data, th=0.001):
    S = np.linalg.svd(data)
    S = np.power(S, 2)
    nse = np.zeros(32)

    for t in range(32):
        S_t = np.power(S, t)
        P = np.divide(S_t, np.sum(S_t, axis=0))
        nse[t] = np.sum(P[np.where(P > th)[0]])

